==> rhgitaly-1.1.3dev0.tgz: rhgitaly-1.1.3dev0/rust/rhgitaly/src/glob.rs <==
// Mostly taken from hg-core/src/matchers.rs
//
// Copyright 2019 Raphaël Gomès <rgomes@octobus.net>
// Copyright 2024 Georges Racinet <georges.racinet@octobus.net> for the minor adaptations
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
use hg::utils::SliceExt;
use lazy_static::lazy_static;
use regex::bytes::Regex;
use std::io::Write;

lazy_static! {
    static ref RE_ESCAPE: Vec<Vec<u8>> = {
        let mut v: Vec<Vec<u8>> = (0..=255).map(|byte| vec![byte]).collect();
        let to_escape = b"()[]{}?*+-|^$\\.&~#\t\n\r\x0b\x0c";
        for byte in to_escape {
            v[*byte as usize].insert(0, b'\\');
        }
        v
    };
}

/// These are matched in order
const GLOB_REPLACEMENTS: &[(&[u8], &[u8])] =
    &[(b"*/", b"(?:.*/)?"), (b"*", b".*"), (b"", b"[^/]*")];

/// Convert a glob pattern into a regular expression
///
/// Taken from [`hg::matchers::re_matcher`] with the following changes:
///
///   - we don't need the multithread optimizations, hence to enclose with thread-locals
///     in `hg::matchers::RegexMatcher`.
///   - we don't need huge regexps, as the main use case in RHGitaly is `fnmatch` on path segments.
#[allow(rustdoc::broken_intra_doc_links)]
pub fn glob_to_regex(pat: &[u8]) -> Regex {
    let rx_pat = glob_to_regex_bytes_pattern(pat);
    let pattern = &rx_pat;

    // The `regex` crate adds `.*` to the start and end of expressions if there
    // are no anchors, so add the start anchor.
    let mut escaped_bytes = vec![b'^', b'(', b'?', b':'];
    for byte in pattern {
        if *byte > 127 {
            write!(escaped_bytes, "\\x{:x}", *byte).unwrap();
        } else {
            escaped_bytes.push(*byte);
        }
    }
    escaped_bytes.push(b')');

    // Avoid the cost of UTF8 checking
    //
    // # Safety
    // This is safe because we escaped all non-ASCII bytes.
    let pattern_string = unsafe { String::from_utf8_unchecked(escaped_bytes) };
    regex::bytes::RegexBuilder::new(&pattern_string)
        .unicode(false)
        .build()
        .unwrap() // TODO unwrap. In theory (must be checked) we cannot make a wrong regexp pattern
}

/// Transforms a glob pattern into a regex
/// Copied over from hg::filepatterns where it is unfortunately private
fn glob_to_regex_bytes_pattern(pat: &[u8]) -> Vec<u8> {
    let mut input = pat;
    let mut res: Vec<u8> = vec![];
    let mut group_depth = 0;

    while let Some((c, rest)) = input.split_first() {
        input = rest;

        match c {
            b'*' => {
                for (source, repl) in GLOB_REPLACEMENTS {
                    if let Some(rest) = input.drop_prefix(source) {
                        input = rest;
                        res.extend(*repl);
                        break;
                    }
                }
            }
            b'?' => res.extend(b"."),
            b'[' => {
                match input.iter().skip(1).position(|b| *b == b']') {
                    None => res.extend(b"\\["),
                    Some(end) => {
                        // Account for the one we skipped
                        let end = end + 1;

                        res.extend(b"[");

                        for (i, b) in input[..end].iter().enumerate() {
                            if *b == b'!' && i == 0 {
                                res.extend(b"^")
                            } else if *b == b'^' && i == 0 {
                                res.extend(b"\\^")
                            } else if *b == b'\\' {
                                res.extend(b"\\\\")
                            } else {
                                res.push(*b)
                            }
                        }
                        res.extend(b"]");
                        input = &input[end + 1..];
                    }
                }
            }
            b'{' => {
                group_depth += 1;
                res.extend(b"(?:")
            }
            b'}' if group_depth > 0 => {
                group_depth -= 1;
                res.extend(b")");
            }
            b',' if group_depth > 0 => res.extend(b"|"),
            b'\\' => {
                let c = {
                    if let Some((c, rest)) = input.split_first() {
                        input = rest;
                        c
                    } else {
                        c
                    }
                };
                res.extend(&RE_ESCAPE[*c as usize])
            }
            _ => res.extend(&RE_ESCAPE[*c as usize]),
        }
    }
    res
}

==> rhgitaly-1.3.0.tgz: rhgitaly-1.3.0/rust/rhgitaly/src/glob.rs <==
// Mostly taken from hg-core/src/matchers.rs
//
// Copyright 2019 Raphaël Gomès <rgomes@octobus.net>
// Copyright 2024 Georges Racinet <georges.racinet@octobus.net> for the minor adaptations
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
use hg::utils::SliceExt;
use lazy_static::lazy_static;
use regex::bytes::Regex;
use std::io::Write;

lazy_static! {
    static ref RE_ESCAPE: Vec<Vec<u8>> = {
        let mut v: Vec<Vec<u8>> = (0..=255).map(|byte| vec![byte]).collect();
        let to_escape = b"()[]{}?*+-|^$\\.&~#\t\n\r\x0b\x0c";
        for byte in to_escape {
            v[*byte as usize].insert(0, b'\\');
        }
        v
    };
}

/// These are matched in order
const GLOB_REPLACEMENTS: &[(&[u8], &[u8])] =
    &[(b"*/", b"(?:.*/)?"), (b"*", b".*"), (b"", b"[^/]*")];

/// Take from hg::matchers (we don't need the multithread optimizations, nor huge regexps
/// main use cases in RHGitaly is fnmatch on path segments.
pub fn glob_to_regex(pat: &[u8]) -> Regex {
    let rx_pat = glob_to_regex_bytes_pattern(pat);
    let pattern = &rx_pat;

    // The `regex` crate adds `.*` to the start and end of expressions if there
    // are no anchors, so add the start anchor.
    let mut escaped_bytes = vec![b'^', b'(', b'?', b':'];
    for byte in pattern {
        if *byte > 127 {
            write!(escaped_bytes, "\\x{:x}", *byte).unwrap();
        } else {
            escaped_bytes.push(*byte);
        }
    }
    escaped_bytes.push(b')');

    // Avoid the cost of UTF8 checking
    //
    // # Safety
    // This is safe because we escaped all non-ASCII bytes.
    let pattern_string = unsafe { String::from_utf8_unchecked(escaped_bytes) };
    regex::bytes::RegexBuilder::new(&pattern_string)
        .unicode(false)
        .build()
        .unwrap() // TODO unwrap. In theory (must be checked) we cannot make a wrong regexp pattern
}

/// Transforms a glob pattern into a regex
/// Copied over from hg::filepatterns where it is unfortunately private
fn glob_to_regex_bytes_pattern(pat: &[u8]) -> Vec<u8> {
    let mut input = pat;
    let mut res: Vec<u8> = vec![];
    let mut group_depth = 0;

    while let Some((c, rest)) = input.split_first() {
        input = rest;

        match c {
            b'*' => {
                for (source, repl) in GLOB_REPLACEMENTS {
                    if let Some(rest) = input.drop_prefix(source) {
                        input = rest;
                        res.extend(*repl);
                        break;
                    }
                }
            }
            b'?' => res.extend(b"."),
            b'[' => {
                match input.iter().skip(1).position(|b| *b == b']') {
                    None => res.extend(b"\\["),
                    Some(end) => {
                        // Account for the one we skipped
                        let end = end + 1;

                        res.extend(b"[");

                        for (i, b) in input[..end].iter().enumerate() {
                            if *b == b'!' && i == 0 {
                                res.extend(b"^")
                            } else if *b == b'^' && i == 0 {
                                res.extend(b"\\^")
                            } else if *b == b'\\' {
                                res.extend(b"\\\\")
                            } else {
                                res.push(*b)
                            }
                        }
                        res.extend(b"]");
                        input = &input[end + 1..];
                    }
                }
            }
            b'{' => {
                group_depth += 1;
                res.extend(b"(?:")
            }
            b'}' if group_depth > 0 => {
                group_depth -= 1;
                res.extend(b")");
            }
            b',' if group_depth > 0 => res.extend(b"|"),
            b'\\' => {
                let c = {
                    if let Some((c, rest)) = input.split_first() {
                        input = rest;
                        c
                    } else {
                        c
                    }
                };
                res.extend(&RE_ESCAPE[*c as usize])
            }
            _ => res.extend(&RE_ESCAPE[*c as usize]),
        }
    }
    res
}
